"""Formatting of expressions for printing."""
import functools as _ft
import math
import textwrap as _tw
import typing as _ty

import parstools._lex as _lex

import tla._ast
import tla._langdef as _opdef


LINE_WIDTH = 80
INDENT_WIDTH = 4
_INDENT = INDENT_WIDTH * '\x20'
_TLA_OPERATOR_FIXITY = _opdef.TLA_OPERATOR_FIXITY


def _make_op_to_fixity(
        ) -> dict:
    """Map operator lexemes to fixity."""
    op_to_fixity = dict()
    items = _TLA_OPERATOR_FIXITY.items()
    for fixity, operators in items:
        for op in operators:
            op_to_fixity[op] = fixity
    return op_to_fixity


_OP_TO_FIXITY: _ty.Final = _make_op_to_fixity()
_nodes = tla._ast.make_nodes()


def pformat_ast(
        tree:
            tuple,
        width:
            int=LINE_WIDTH
        ) -> str:
    """Return formatted tree."""
    if isinstance(tree, str):
        return tree
    rec = _ft.partial(
        pformat_ast,
        width=width)
    enum_name = tla._ast.enum_to_class(
        tree.symbol.name)
    match enum_name:
        case 'Axiom':
            expr = rec(tree.expression)
            match tree.axiom_keyword:
                case None:
                    keyword = 'AXIOM'
                case (
                        'AXIOM' |
                        'ASSUME' |
                        'ASSUMPTION'):
                    keyword = str(
                        tree.axiom_keyword)
                case _:
                    raise ValueError(
                        tree.axiom_keyword)
            if tree.name is None:
                axiom = keyword
            else:
                axiom = f'{keyword} {tree.name} =='
            return _glue_prefix_box(
                axiom,
                f'\n{expr}',
                indent_width=INDENT_WIDTH)
        case 'BooleanLiteral':
            match tree.value:
                case 'FALSE':
                    if width < len('FALSE'):
                        raise AssertionError(width)
                    return 'FALSE'
                case 'TRUE':
                    if width < len('TRUE'):
                        raise AssertionError(width)
                    return 'TRUE'
                case _:
                    raise ValueError(
                        tree.value)
        case 'By':
            if tree.facts is None:
                facts = list()
            else:
                facts = list(map(
                    rec, tree.facts))
            if tree.names is None:
                names = list()
            else:
                names = list(map(
                    rec, tree.names))
            if tree.only:
                only = ' ONLY'
            else:
                only = ''
            prefix = f'{_INDENT}BY{only}'
            return _format_usable(
                facts, names, prefix,
                width=width)
        case 'Cases':
            cases = list()
            for case_item in tree.cases:
                case_str = rec(case_item)
                cases.append(case_str)
            cases_str = ' [] '.join(cases)
            res = f'CASE {cases_str}'
            indent_width = 2
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                cases_str = '\n[] '.join(cases)
                res = _glue_prefix_box(
                    'CASE ',
                    cases_str,
                    indent_width=indent_width)
            if tree.other is None:
                return res
            other = rec(tree.other)
            if res_width > width:
                return _glue_prefix_box(
                    res,
                    f'\n[] OTHER -> {other}',
                    indent_width=indent_width)
            return f'{res} [] OTHER -> {other}'
        case 'CaseItem':
            pred = rec(tree.predicate)
            expr = rec(tree.expression)
            return f'{pred} -> {expr}'
        case 'Choose':
            pred = rec(tree.predicate)
            decl = rec(tree.declaration)
            choose = f'CHOOSE {decl}:  '
            return _concatenate_boxes(
                choose, pred, width)
        case 'Constants':
            if not tree.names:
                raise ValueError(
                    tree.names)
            decls = list(map(
                rec, tree.names))
            decls_str = ',\n'.join(decls)
            # decls_str = _tw.fill(
            #     decls_str,
            #     width=width - len('CONSTANT '))
            return _glue_prefix_box(
                'CONSTANT',
                '\n' + decls_str,
                indent_width=INDENT_WIDTH)
        case 'Define':
            indent_width = 2 * INDENT_WIDTH
            width -= 3 * INDENT_WIDTH
            defns_str = _format_definitions(
                tree.definitions,
                width=width)
            return _glue_prefix_box(
                'DEFINE',
                f'\n{defns_str}',
                indent_width=indent_width)
        case 'Except':
            func = rec(tree.function)
            mods = list()
            for eq in tree.changes:
                mod = rec(eq)
                mods.append(mod)
            mods_str = ', '.join(mods)
            mods_width, _ = _box_dimensions(
                mods_str)
            if mods_width > width:
                mods_str = ',\n'.join(mods)
                return _glue_prefix_box(
                    f'[{func} EXCEPT',
                    f'\n{mods_str}',
                    indent_width=INDENT_WIDTH)
            return f'[{func} EXCEPT {mods_str}]'
        case 'Fairness':
            op = tree.operator
            if op not in ('WF_', 'SF_'):
                raise ValueError(
                    tree.operator)
            subscript = rec(tree.subscript)
            action = rec(tree.action)
            return f'{op}{subscript}({action})'
        case 'Field':
            expr = rec(tree.expression)
            name = tree.name
            # parentheses needed ?
            needs_parentheses = (
                '\x20' in expr and
                not expr.endswith(')') and
                not expr.endswith(']') and
                not expr.startswith('{') and
                not expr.startswith('<<'))
            if needs_parentheses:
                raise ValueError(
                    'Field expression needs '
                    f'parentheses: {tree}')
            return f'{expr}.{name}'
        case 'FloatNumeral':
            integer, mantissa = tree.value.split('.')
            if not mantissa:
                raise ValueError(tree.value)
            return tree.value
        case 'Function':
            new_width = width - INDENT_WIDTH
            rec = _ft.partial(
                pformat_ast,
                width=new_width)
            exprs = list(map(
                rec, tree.declaration))
            bounds_str = ', '.join(exprs)
            expr = rec(tree.value)
            start = _glue_prefix_box(
                '[',
                bounds_str,
                indent_width=INDENT_WIDTH)
            # combine with expr
            return _concatenate_boxes(
                f'{start} |-> ',
                f'{expr}]',
                width,
                INDENT_WIDTH,
                INDENT_WIDTH)
        case 'FunctionApplication':
            func = rec(tree.function)
            args_list = ', '.join(map(
                rec, tree.arguments))
            needs_parentheses = (
                '\x20' in func and
                not func.endswith(')') and
                    # includes nonfix-operator
                    # application
                not func.endswith(']') and
                not func.startswith('{') and
                not func.startswith('<<'))
            if needs_parentheses:
                print(
                    'Function may need parentheses '
                    'in function application: '
                    f'{func}')
            return f'{func}[{args_list}]'
        case 'FunctionChange':
            args = list()
            for arg in tree.item:
                if isinstance(arg, str):
                    args.append(
                        f'.{arg}')
                elif isinstance(arg, list):
                    exprs = ', '.join(map(
                        rec, arg))
                    args.append(
                        f'[{exprs}]')
                else:
                    raise AssertionError(arg)
            expr = rec(tree.expression)
            args_str = ''.join(args)
            return f'!{args_str} = {expr}'
        case 'Have':
            expr = rec(tree.expression)
            return _glue_prefix_box(
                'HAVE ',
                expr,
                indent_width=2 * INDENT_WIDTH)
        case 'Hide':
            if tree.facts is None:
                facts = list()
            else:
                facts = list(map(
                    rec, tree.facts))
            if tree.names is None:
                names = list()
            else:
                names = list(map(
                    rec, tree.names))
            return _format_usable(
                facts, names,
                'HIDE',
                width=width)
        case 'If':
            pred = rec(tree.predicate)
            then = rec(tree.then)
            else_ = rec(tree.else_)
            res = (
                f'IF {pred} THEN {then} '
                f'ELSE {else_}')
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                res = _glue_prefix_box(
                    f'IF {pred}',
                    f'\nTHEN {then}',
                    indent_width=INDENT_WIDTH)
                res = _glue_prefix_box(
                    res,
                    f'\nELSE {else_}',
                    indent_width=INDENT_WIDTH)
            return res
        case 'IfItem':
            pred = rec(tree.predicate)
            expr = rec(tree.expression)
            return f'ELIF {pred} THEN {expr}'
        case 'Ifs':
            first, *rest = tree.cases
            else_ = rec(tree.other)
            pred = rec(first.predicate)
            expr = rec(first.expression)
            # elifs
            elifs = '\n'.join(map(rec, rest))
            elifs = f'\n{elifs}\nELSE {else_}'
            # combine
            res = f'IF {pred} THEN {expr}'
            return _glue_prefix_box(
                res,
                elifs,
                indent_width=INDENT_WIDTH)
        case 'Instance':
            inst = f'INSTANCE {tree.name}'
            if tree.local:
                inst = f'LOCAL {inst}'
            if tree.with_substitution is None:
                return inst
            substitutions = list(map(
                rec, tree.with_substitution))
            subs_str = ', '.join(substitutions)
            if subs_str:
                inst = f'{inst} WITH {subs_str}'
            return inst
        case 'IntegralNumeral':
            return f'{tree.value}'
        case 'KeyBound':
            expr = rec(tree.bound)
            return f'{tree.name}: {expr}'
        case 'Lambda':
            args = list(map(
                rec, tree.parameters))
            expr = rec(tree.expression)
            args = _format_lambda_signature(
                args, width)
            s = _glue_prefix_box(
                'LAMBDA ', args)
            return _glue_prefix_box(
                f'{s}:  ', expr)
        case 'Let':
            expr = rec(tree.expression)
            defns_str = '\n'.join(map(
                rec, tree.definitions))
            indent_width = len('LET ')
            res = _glue_prefix_box(
                '\nLET',
                f'\n{defns_str}',
                indent_width=indent_width)
            res += '\nIN'
            return _glue_prefix_box(
                res,
                f'\n{expr}',
                indent_width=indent_width)
        case 'Module':
            # module start
            n = (width - len(tree.name)
                - len(' MODULE  ')) / 2
            left = math.floor(n)
            right = math.ceil(n)
            title = ''.join([
                '-' * left,
                ' MODULE ',
                tree.name, ' ',
                '-' * right])
            # `EXTENDS` statement
            if tree.extendees is None:
                extends = None
            else:
                exts = [
                    _INDENT + rec(ext)
                    for ext in tree.extendees]
                exts_str = ',\n'.join(exts)
                extends = f'EXTENDS\n{exts_str}'
            # content
            units = '\n'.join(map(
                rec, tree.units))
            endline = width * '='
            if extends is None:
                parts = [
                    title, units, endline]
            else:
                parts = [
                    title, extends,
                    units, endline]
            return '\n'.join(parts)
        case 'ModuleName':
            return f'MODULE {tree.name}'
        case 'Obvious':
            return f'{_INDENT}OBVIOUS'
        case 'Omitted':
            return f'{_INDENT}OMITTED'
        case 'OperatorApplication':
            if isinstance(tree.operator, str):
                op = tree.operator
            elif tree.operator is None:
                # in subexpression reference
                op = ''
            else:
                op = rec(tree.operator)
            if tree.arguments is None:
                return op
            args = list(map(
                rec, tree.arguments))
            n_args = len(args)
            # pre-defined symbol ?
            if op in _OP_TO_FIXITY:
                fixity = _OP_TO_FIXITY[op]
            else:
                fixity = None
            if op == '-':
                match n_args:
                    case 1:
                        fixity = 'before'
                    case 2:
                        fixity = 'left'
                    case _:
                        raise ValueError(
                            op, n_args)
            if op == '-.':
                # op = '-'
                fixity = 'before'
            if op == '\\X':
                return f' {op} '.join(args)
            if fixity is None:
                pass
            elif n_args == 2:
                if fixity not in (
                        'infix', 'left',
                        'between', 'right'):
                    raise ValueError(
                        'prefix or postfix operator '
                        'applied to two arguments')
                arg0 = args[0]
                # not vertical expression as
                # first argument ?
                if not arg0.endswith('\n'):
                    arg0 = f'{arg0} '
                res = f'{arg0}{op} {args[1]}'
                res_width, _ = _box_dimensions(res)
                if res_width > width:
                    res = _glue_prefix_box(
                        args[0],
                        op)
                    res = _glue_prefix_box(
                        f'{res}\n',
                        args[1])
                return res
            elif n_args != 1:
                raise ValueError(
                    f'operator `{op}` applied to '
                    f'{n_args} arguments (expected 1)')
            elif fixity in ('prefix', 'before'):
                return f'{op} {args[0]}'
            elif fixity in ('postfix', 'after'):
                return f'{args[0]}{op}'
            # nonfix-operator application
            assert fixity is None
            args_str = ', '.join(args)
            res = f'{op}({args_str})'
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                args_str = ',\n'.join(
                    f'    {arg}'
                    for arg in args)
                res = f'{op}(\n{args_str})'
            return res
        case 'OperatorDeclaration':
            expr = rec(tree.name)
                # includes arity declaration
            match tree.level:
                case None:
                    if not tree.new_keyword:
                        raise ValueError(tree)
                    level = ''
                case (
                        'VARIABLE' |
                        'CONSTANT'|
                        'STATE' |
                        'ACTION' |
                        'TEMPORAL'):
                    level = tree.level
                case _:
                    raise ValueError(tree.level)
            if tree.new_keyword:
                new = 'NEW '
            else:
                new = ''
            return f'{new}{level} {expr}'
        case 'OperatorDefinition':
            if tree.function:
                # f[...] == ...
                definiens = rec(tree.definiens.value)
                decl = tree.definiens.declaration
                sig = _nodes.FunctionApplication(
                    function=tree.name,
                    arguments=decl)
            else:
                # f == ...
                # f(...) == ...
                definiens = rec(tree.definiens)
                sig = _nodes.OperatorApplication(
                    operator=tree.name,
                    arguments=tree.arity)
            sig = rec(sig)
            if tree.local:
                local = 'LOCAL '
            else:
                local = ''
            res = f'{local}{sig} == '
            res2 = _glue_prefix_box(
                res,
                definiens,
                indent_width=INDENT_WIDTH)
            res2_width, _ = _box_dimensions(res2)
            if res2_width > width:
                if not definiens.startswith('\n'):
                    definiens = f'\n{definiens}'
                res2 = _glue_prefix_box(
                    res,
                    definiens,
                    indent_width=INDENT_WIDTH)
            return res2
        case 'ParameterDeclaration':
            if tree.arity is None:
                return str(tree.name)
            else:
                arity = ['_'] * tree.arity
                arity = ', '.join(arity)
                return f'{tree.name}({arity})'
        case 'Parentheses':
            expr = rec(tree.expression)
            return f'({expr})'
        case 'Pick':
            # format parts
            exprs = list(map(
                rec, tree.declarations))
            decls = ', '.join(exprs)
            pred = rec(tree.predicate)
            # glue
            pick = f'PICK {decls}:  '
            pick_width, _ = _box_dimensions(pick)
            if pick_width > width:
                pick = _glue_prefix_box(
                    f'PICK',
                    f'\n{decls}:  ')
            res = _glue_prefix_box(
                pick,
                pred,
                indent_width=INDENT_WIDTH)
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                res = _glue_prefix_box(
                    pick,
                    f'\n{pred}',
                    indent_width=INDENT_WIDTH)
            return res
        case 'Proof':
            new_width = width - INDENT_WIDTH
            rec = _ft.partial(
                pformat_ast,
                width=new_width)
            steps = list(map(
                rec, tree.steps))
            proof = '\n'.join(steps)
            if tree.proof_keyword:
                proof = f'PROOF\n{proof}'
            return _tw.indent(
                proof,
                prefix=_INDENT)
        case 'ProofCase':
            expr = rec(tree.expression)
            return f'CASE {expr}'
        case 'ProofStep':
            step_number = tree.name
            if tree.main == 'QED':
                main = tree.main
            else:
                main = rec(tree.main)
                assert main is not None
            pf_goal = _glue_prefix_box(
                f'{step_number} ', main)
            if tree.proof:
                leaf_proof = rec(tree.proof)
            else:
                leaf_proof = ''
            if tree.proof_keyword:
                pf = f'{_INDENT}PROOF\n'
            else:
                pf = ''
            return _glue_prefix_box(
                pf_goal,
                f'\n{pf}{leaf_proof}',
                indent_width=0)
        case 'Quantification':
            quantifier = tree.quantifier
            if quantifier not in (r'\A', r'\E'):
                raise ValueError(
                    tree.quantifier)
            # declarations
            decls_width = (
                width
                - len(quantifier)
                - len(' :'))
            rec = _ft.partial(
                pformat_ast,
                width=decls_width)
            decls = ', '.join(map(
                rec, tree.declarations))
            # expr
            expr_width = width - INDENT_WIDTH
            pred = pformat_ast(
                tree.predicate,
                width=expr_width)
            # combine
            res = _concatenate_boxes(
                f'{quantifier} ',
                f'{decls}:  ',
                width,
                INDENT_WIDTH,
                INDENT_WIDTH)
            res_width, _ = _box_dimensions(res)
            return _glue_prefix_box(
                res,
                pred,
                indent_width=res_width)
        case 'QuantifierWitness':
            exprs = ', '.join(map(
                rec, tree.expressions))
            return f'WITNESS {exprs}'
        case 'Record':
            pairs = list()
            for name, expr in tree.key_values:
                name = rec(name)
                expr = rec(expr)
                pair = f'{name} |-> {expr}'
                pairs.append(pair)
            items_str = ', '.join(pairs)
            return f'[{items_str}]'
        case 'RecursiveOperators':
            names = ', '.join(map(
                rec, tree.names))
            return f'RECURSIVE {names}'
        case 'Sequent':
            new_width = width - INDENT_WIDTH
            rec = _ft.partial(
                pformat_ast,
                width=new_width)
            assumptions = list(map(
                rec, tree.assume))
            goal = rec(tree.prove)
            res = _glue_prefix_box(
                'ASSUME',
                '\n' + _join_boxes_sep(
                    assumptions,
                    sep=', '),
                indent_width=INDENT_WIDTH)
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                res = _glue_prefix_box(
                    'ASSUME',
                    '\n' + ',\n'.join(
                        assumptions),
                    indent_width=INDENT_WIDTH)
            return _glue_prefix_box(
                f'{res}\nPROVE',
                f'\n{goal}',
                indent_width=INDENT_WIDTH)
        case 'SetComprehension':
            decls = ', '.join(map(
                rec, tree.declarations))
            expr = rec(tree.item)
            res = _glue_prefix_box(
                '{' + expr + ':  ',
                decls + '}')
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                res = _glue_prefix_box(
                    '{' + expr + ':',
                    '\n' + decls + '}',
                    indent_width=INDENT_WIDTH)
            return res
        case 'SetEnumeration':
            exprs = list(map(
                rec, tree.items))
            exprs_str = ', '.join(exprs)
            exprs_width, _ = _box_dimensions(
                exprs_str)
            if exprs_width > width:
                exprs_str = ',\n'.join(exprs)
            return '{' + exprs_str + '}'
        case 'SetOfBooleans':
            return 'BOOLEAN'
        case 'SetOfFunctions':
            domain = rec(tree.domain)
            codomain = rec(tree.codomain)
            return f'[{domain} -> {codomain}]'
        case 'SetOfRecords':
            items = ', '.join(map(
                rec, tree.key_bounds))
            return f'[{items}]'
        case 'SetOfStrings':
            return 'STRING'
        case 'SetSlice':
            # can be tuply
            decl = rec(tree.declaration)
            pred = rec(tree.predicate)
            decl_str = '{' + f'{decl}:  '
            res = _glue_prefix_box(
                decl_str,
                pred + '}',
                indent_width=INDENT_WIDTH)
            res_width, _ = _box_dimensions(res)
            if res_width > width:
                res = _glue_prefix_box(
                    decl_str,
                    '\n' + pred + '}',
                    indent_width=INDENT_WIDTH)
            return res
        case 'StringLiteral':
            return f'{tree.value}'
        case 'SubexpressionReference':
            return '!'.join(map(
                rec, tree.items))
        case 'SubscriptedAction':
            action = rec(tree.action)
            subscript = rec(tree.subscript)
            match tree.operator:
                case '[':
                    return f'[{action}]_{subscript}'
                case '<<':
                    return f'<<{action}>>_{subscript}'
                case _:
                    raise ValueError(
                        tree.operator)
        case 'Suffices':
            pred = rec(tree.predicate)
            return _glue_prefix_box(
                'SUFFICES',
                f'\n{pred}',
                indent_width=2 * INDENT_WIDTH)
        case 'Take':
            decls = ', '.join(map(
                rec, tree.declarations))
            return _glue_prefix_box(
                'TAKE ',
                decls,
                indent_width=2 * INDENT_WIDTH)
        case 'TemporalQuantification':
            quantifier = tree.quantifier
            if quantifier not in (r'\AA', r'\EE'):
                raise ValueError(
                    tree.quantifier)
            names = ', '.join(map(
                rec, tree.declarations))
            pred_width = width - INDENT_WIDTH
            pred = pformat_ast(
                tree.predicate,
                width=pred_width)
            qprefix = _concatenate_boxes(
                f'{quantifier} ',
                f'{names}:  ',
                width,
                INDENT_WIDTH,
                INDENT_WIDTH)
            return _concatenate_boxes(
                qprefix, pred, width,
                INDENT_WIDTH, INDENT_WIDTH)
        case 'Theorem':
            goal = pformat_ast(
                tree.goal,
                width=width - INDENT_WIDTH)
            theorem_keyword = tree.theorem_keyword
            if theorem_keyword not in (
                    'PROPOSITION',
                    'LEMMA',
                    'THEOREM',
                    'COROLLARY'):
                raise ValueError(
                    tree.theorem_keyword)
            if tree.name is None:
                theorem = theorem_keyword
            else:
                theorem = (
                    f'{theorem_keyword} '
                    f'{tree.name} ==')
            res = _glue_prefix_box(
                theorem,
                f'\n{goal}',
                indent_width=INDENT_WIDTH)
            if tree.proof is None:
                return res
            proof_str = rec(tree.proof)
            proof_str = _tw.dedent(proof_str)
            return _glue_prefix_box(
                res,
                f'\n{proof_str}',
                indent_width=0)
        case 'Tuple':
            items = list(map(
                rec, tree.items))
            items_str = ', '.join(items)
            tpl = f'<<{items_str}>>'
            tpl_width, _ = _box_dimensions(tpl)
            if tpl_width > width:
                tpl = ',\n'.join(items)
                tpl = _glue_prefix_box(
                    '<<',
                    f'\n{tpl}>>')
            return tpl
        case 'Use':
            if tree.facts is None:
                facts = list()
            else:
                facts = list(map(
                    rec, tree.facts))
            if tree.names is None:
                names = list()
            else:
                names = list(map(
                    rec, tree.names))
            if tree.only:
                only = ' ONLY'
            else:
                only = ''
            prefix = f'USE{only}'
            return _format_usable(
                facts, names, prefix,
                width=width)
        case 'Variables':
            if not tree.names:
                raise ValueError(tree.names)
            vrs_str = ',\n'.join(map(
                rec, tree.names))
            return _glue_prefix_box(
                'VARIABLE',
                '\n' + vrs_str,
                indent_width=INDENT_WIDTH)
        case 'VerticalList':
            if not tree.arguments:
                raise ValueError(
                    tree.arguments)
            op = tree.operator
            # assert same operator
            for other in tree.arguments:
                if other.operator == op:
                    continue
                raise ValueError(
                    other.operator, op)
            # format
            new_width = width - INDENT_WIDTH
            blocks = list()
            for expr in tree.arguments:
                expr = rec(
                    expr,
                    width=new_width)
                block = _glue_prefix_box(
                    f'{op} ',
                    expr)
                blocks.append(block)
            # combine
            return '\n' + '\n'.join(blocks) + '\n'
        case 'VerticalItem':
            return rec(tree.expression)
        case 'WithInstantiation':
            name = tree.name
            expr = rec(tree.expression)
            return f'{name} <- {expr}'
        case _:
            raise ValueError(
                tree.symbol.name)


def _box_dimensions(
        string:
            str
        ) -> tuple[
            int,
            int]:
    r"""Return width, height of `string`.

    Width is the number of characters in the
    longest line. Height is the number of lines,
    which includes a newline (`\n`) at the end.
    """
    lines = string.split('\n')
        # counts the `\n` at end
    widths = [
        len(line)
        for line in lines]
    width = max(widths)
    height = len(lines)
    return (
        width,
        height)


def _glue_prefix_box(
        prefix:
            str,
        box:
            str,
        indent_width:
            int |
            None=None
        ) -> str:
    """Concatenate strings with proper indentation."""
    lines = box.split('\n')
    if indent_width is None:
        prefix_lines = prefix.split('\n')
        indent_width = len(prefix_lines[-1])
    indent = indent_width * '\x20'
    res_lines = [prefix + lines[0]]
    res_lines.extend(
        indent + line
        for line in lines[1:])
    return '\n'.join(res_lines)


def _concatenate_boxes(
        box1:
            str,
        box2:
            str,
        width:
            int,
        indent_width:
            int=INDENT_WIDTH,
        alt_indent_width:
            int=INDENT_WIDTH
        ) -> str:
    """Concatenate `box1` and `box2` within `width`.

    Introduce a newline if the result would be
    wider than `width`. Use `indent_width` to indent
    `box2` if concatenated without newline,
    otherwise `alt_indent_width`.
    """
    res = _glue_prefix_box(
        box1, box2, indent_width)
    res_width, _ = _box_dimensions(res)
    if res_width <= width:
        return res
    return _glue_prefix_box(
        box1,
        f'\n{box2}',
        alt_indent_width)


def _join_boxes_sep(
        boxes:
            list,
        sep:
            str
        ) -> str:
    """Join boxes by iterative gluing."""
    res = ''
    for box in boxes:
        if res:
            res += sep
        res = _glue_prefix_box(
            res, box)
    return res


def _format_lambda_signature(
        args:
            list[str],
        width:
            int
        ) -> str:
    """Return `str` of `LAMBDA` signature."""
    res = ', '.join(args)
    res_width, _ = _box_dimensions(res)
    if res_width > width:
        res = ',\n'.join(args)
    return res


def _format_definitions(
        definitions:
            list,
        width:
            int |
            None=None
        ) -> str:
    r"""Return formatted `definitions`.

    Concatenate vertically the string
    representations of `definitions` by
    inserting newlines `\n`.
    """
    defns = list()
    for defn in definitions:
        defn_str = pformat_ast(
            defn,
            width=width)
        defns.append(defn_str)
    return '\n'.join(defns)


def _format_usable(
        facts:
            list[str],
        defs:
            list[str],
        prefix:
            str,
        width:
            int |
            None=None
        ) -> str:
    """Return formatted with `prefix`."""
    # facts
    facts_str = ', '.join(facts)
    # new_width = width - len(prefix) - INDENT_WIDTH
    # facts_str = _tw.fill(
    #     facts_str,
    #     width=new_width)
    if facts:
        res = _glue_prefix_box(
            prefix,
            f' {facts_str}',
            indent_width=INDENT_WIDTH)
    else:
        res = prefix
    # defs
    defs_str = ', '.join(defs)
    # new_width = (
    #     width - len('DEF ') - len(prefix)
    #     - INDENT_WIDTH)
    # defs_str = _tw.fill(
    #     defs_str,
    #     width=new_width)
    # combine
    if defs:
        defs_str = f'DEF {defs_str}'
        res = _glue_prefix_box(
            res,
            f' {defs_str}',
            indent_width=INDENT_WIDTH)
    res_width, _ = _box_dimensions(res)
    if res_width > width:
        res = _glue_prefix_box(
            f'{prefix} ',
            facts_str,
            indent_width=len(prefix) + 1)
        if defs:
            res = _glue_prefix_box(
                res,
                f'\n{defs_str}',
                indent_width=len(prefix) + 1)
    return res


def _print_overwide_lines(
        text:
            str,
        max_width:
            int
        ) -> None:
    """Print lines wider than `max_width`.

    Print 5 lines of context before and after the
    widest lines in `text`. Print the width of the
    widest line in `text`.
    """
    width, _ = _box_dimensions(text)
    print(f'Width of text: {width}')
    lines = text.split('\n')
    for i, line in enumerate(lines):
        if len(line) <= max_width:
            continue
        span = '\n'.join(lines[i - 5: i + 6])
        print(span)


def pformat_tla(
        text:
            str
        ) -> str:
    """Return formatted `text` with comments.

    The comments are inserted again after
    formatting.
    """
    lexer = tla.Lexer()
    tokens = list(lexer.parse(text))
    # tokens of formatted text
    tree = tla.parse(text)
    formatted_text = tla.pformat_ast(tree)
    ptokens = list(lexer.parse(
        formatted_text))
    # copy comments
    ctokens = _zip_tokens(tokens, ptokens)
    ctokens = _reposition_tokens(ctokens)
    return _lex.join_tokens(ctokens)


_COMMENT_TOKENS = {
    'UNILINE_COMMENT',
    'MULTILINE_COMMENT'}


def _zip_tokens(
        tokens:
            list,
        ptokens:
            list
        ) -> list:
    """Copy comments from `tokens` to `ptokens`.

    Comments are copied with relative positions.
    """
    rest = iter(ptokens)
    ctokens = list()
        # tokens with comments inserted
    last_token = _lex.make_none_token()
    for token in tokens:
        if token.symbol in _COMMENT_TOKENS:
            start, row, column = _pos_diff(
                last_token, token)
            ctoken = _lex.Token(
                symbol=token.symbol,
                value=token.value,
                start=start,
                row=row,
                column=column)
            ctokens.append(ctoken)
            last_token = token
            continue
        ptoken = next(rest, None)
        if ptoken is None:
            raise AssertionError(
                'different number of '
                f'non-comment tokens at: {token}')
        if token.symbol != ptoken.symbol:
            raise AssertionError(
                token, ptoken)
        ctokens.append(ptoken)
        last_token = token
    ptoken = next(rest, None)
    if ptoken is not None:
        raise AssertionError(
            'different number of '
            f'non-comment tokens at: {ptoken}')
    return ctokens


def _pos_diff(
        last_token,
        token):
    """Return relative position."""
    span = len(last_token.value)
    start = (
        token.start
        - span
        - last_token.start)
    row = (
        token.row
        - last_token.row)
    # consecutive comments changed to
    # different lines
    comments_same_line = (
        last_token.symbol in
            _COMMENT_TOKENS and
        row == last_token.count('\n'))
    if comments_same_line:
        row += 1
    # same line ?
    if row == 0:
        column = (
            token.column
            - span
            - last_token.column)
    elif row < 0:
        raise AssertionError(
            last_token, token)
    else:
        column = token.column
    return (
        start,
        row,
        column)


def _reposition_tokens(
        tokens):
    """Update token coordinates.

    Based on relative positions and size of
    comments, update the positions of all
    tokens, including of comment tokens.
    """
    row = 0
    row_shift = 0
    ctokens = list()
    last_token = _lex.make_none_token()
    for index, token in enumerate(tokens):
        span = len(last_token.value)
        if token.symbol in _COMMENT_TOKENS:
            row = (
                last_token.row +
                token.row)
            if token.row:
                column = token.column
            else:
                column = (
                    last_token.column +
                    token.column +
                    span)
            n_newlines = token.value.count('\n')
            row_shift += (
                token.row +
                n_newlines)
            # add newline after multiline comments
            # to avoid changing columns within
            # vertical expressions
            end_row = (
                row +
                n_newlines)
            if index + 1 < len(tokens):
                next_token = tokens[index + 1]
            else:
                next_token = None
            add_newline = (
                next_token is not None and
                next_token.symbol not in
                    _COMMENT_TOKENS and
                next_token.row + row_shift == end_row and
                token.symbol == 'MULTILINE_COMMENT')
                    # uniline comments end at a newline
            if add_newline:
                row_shift += 1
        else:
            row = token.row + row_shift
            column = token.column
                # column numbers unchanged to
                # keep alignment of vertical operators
        ctoken = _lex.Token(
            symbol=token.symbol,
            value=token.value,
            start=None,
            row=row,
            column=column)
        ctokens.append(ctoken)
        last_token = ctoken
    return ctokens



"""Tests of module `tla._lex`."""
import pprint
import textwrap as _tw

import tla._lex as _lex


module_example = r'''
comments
---- MODULE example ----
VARIABLE x

\* a one-line comment
\* a nested \* one-line comment

(* This is a multi-line comment. *)
(* This is a multi-line
comment. * (
*)
(* A nested
(* multi-line *)
comment. *)

a == 1.0
b == a + 2

A == x' = x + 1

P == SF_x(A)

THEOREM Thm ==
    ASSUME x = 1
    PROVE x + 1 = 2
PROOF
<1>1. x = 1
    OBVIOUS
<1>2. x + 1 = 1 + 1
    BY <1>1
<1>3. 1 + 1 = 2
    OBVIOUS
<1> QED
    BY <1>2, <1>3
====================
comments
'''


def test_lexing_module_example():
    lexer = _lex.Lexer()
    tokens = lexer.parse(module_example)
    for token in tokens:
        print(token)
    tv = _tokenize(module_example, lexer)
    tv_ = [
        ('NAME', 'comments'),
        ('DASH_LINE', '----'),
        ('MODULE', 'MODULE'),
        ('NAME', 'example'),
        ('DASH_LINE', '----'),
        ('VARIABLE', 'VARIABLE'),
        ('NAME', 'x'),
        ('UNILINE_COMMENT', '\\* a one-line comment\n'),
        ('UNILINE_COMMENT',
         '\\* a nested \\* one-line comment\n'),
        ('MULTILINE_COMMENT',
         '(* This is a multi-line comment. *)'),
        ('MULTILINE_COMMENT',
        '(* This is a multi-line\ncomment. * (\n*)'),
        ('MULTILINE_COMMENT',
        '(* A nested\n(* multi-line *)\ncomment. *)'),
        ('NAME', 'a'),
        ('DOUBLE_EQ', '=='),
        ('FLOAT', '1.0'),
        ('NAME', 'b'),
        ('DOUBLE_EQ', '=='),
        ('NAME', 'a'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '2'),
        ('NAME', 'A'),
        ('DOUBLE_EQ', '=='),
        ('NAME', 'x'),
        ('PRIME', "'"),
        ('EQ', '='),
        ('NAME', 'x'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '1'),
        ('NAME', 'P'),
        ('DOUBLE_EQ', '=='),
        ('SF_', 'SF_'),
        ('NAME', 'x'),
        ('LPAREN', '('),
        ('NAME', 'A'),
        ('RPAREN', ')'),
        ('THEOREM', 'THEOREM'),
        ('NAME', 'Thm'),
        ('DOUBLE_EQ', '=='),
        ('ASSUME', 'ASSUME'),
        ('NAME', 'x'),
        ('EQ', '='),
        ('DECIMAL_INTEGER', '1'),
        ('PROVE', 'PROVE'),
        ('NAME', 'x'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '1'),
        ('EQ', '='),
        ('DECIMAL_INTEGER', '2'),
        ('PROOF', 'PROOF'),
        ('STEP_NUMBER', '<1>1.'),
        ('NAME', 'x'),
        ('EQ', '='),
        ('DECIMAL_INTEGER', '1'),
        ('OBVIOUS', 'OBVIOUS'),
        ('STEP_NUMBER', '<1>2.'),
        ('NAME', 'x'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '1'),
        ('EQ', '='),
        ('DECIMAL_INTEGER', '1'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '1'),
        ('BY', 'BY'),
        ('STEP_NUMBER', '<1>1'),
        ('STEP_NUMBER', '<1>3.'),
        ('DECIMAL_INTEGER', '1'),
        ('PLUS', '+'),
        ('DECIMAL_INTEGER', '1'),
        ('EQ', '='),
        ('DECIMAL_INTEGER', '2'),
        ('OBVIOUS', 'OBVIOUS'),
        ('STEP_NUMBER', '<1>'),
        ('QED', 'QED'),
        ('BY', 'BY'),
        ('STEP_NUMBER', '<1>2'),
        ('COMMA', ','),
        ('STEP_NUMBER', '<1>3'),
        ('EQ_LINE', '===================='),
        ('NAME', 'comments'),
        ]
    assert tv == tv_, tv


def test_lexing_nested_multiline_comments():
    lexer = _lex.Lexer()
    source = '''
        ------ MODULE nested_multiline -----
        (* outer comment
            (* inner comment *)
        *)
        =====
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        ('DASH_LINE', '------'),
        ('MODULE', 'MODULE'),
        ('NAME', 'nested_multiline'),
        ('DASH_LINE', '-----'),
        ('MULTILINE_COMMENT',
        '''(* outer comment
            (* inner comment *)
        *)'''),
        ('EQ_LINE', '=====')]
    assert tv == tv_, tv


def test_lexing_nested_uniline_comments():
    lexer = _lex.Lexer()
    source = r'''
        ---- MODULE nested_uniline -----
        \* outer comment \* inner comment
        ====
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        ('DASH_LINE', '----'),
        ('MODULE', 'MODULE'),
        ('NAME', 'nested_uniline'),
        ('DASH_LINE', '-----'),
        ('UNILINE_COMMENT',
            r'\* outer comment \* '
            'inner comment\n'),
        ('EQ_LINE', '====')]
    assert tv == tv_, tv


def test_uniline_inside_multiline_comments():
    lexer = _lex.Lexer()
    source = r'''
        ----- MODULE nested_mixed ----
        (* multiline comment \* uniline comment *)
        ======
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        ('DASH_LINE', '-----'),
        ('MODULE', 'MODULE'),
        ('NAME', 'nested_mixed'),
        ('DASH_LINE', '----'),
        ('MULTILINE_COMMENT',
            '(* multiline comment '
            r'\* uniline comment *)'),
        ('EQ_LINE', '======')]
    assert tv == tv_, tv


def test_multiline_inside_uniline_comments():
    lexer = _lex.Lexer()
    source = r'''
        ---- MODULE nested_mixed ----
        \* uniline comment (* multiline comment *)
        ====
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        ('DASH_LINE', '----'),
        ('MODULE', 'MODULE'),
        ('NAME', 'nested_mixed'),
        ('DASH_LINE', '----'),
        ('UNILINE_COMMENT',
            r'\* uniline comment '
            '(* multiline comment *)\n'),
        ('EQ_LINE', '====')]
    assert tv == tv_, tv


def test_string_meta_sequences():
    lexer = _lex.Lexer()
    source = r'''
        ---- MODULE string_meta ----
        quote == "\""
        literal == "\\"
        tab == "\t"
        line_feed == "\n"
        form_feed == "\f"
        carriage_return == "\r"
        ====
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        ('DASH_LINE', '----'),
        ('MODULE', 'MODULE'),
        ('NAME', 'string_meta'),
        ('DASH_LINE', '----'),
        # quote == "\""
        ('NAME', 'quote'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\""'),
        # literal == "\\"
        ('NAME', 'literal'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\\"'),
        # tab == "\t"
        ('NAME', 'tab'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\t"'),
        # line_feed == "\n"
        ('NAME', 'line_feed'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\n"'),
        # form_feed == "\f"
        ('NAME', 'form_feed'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\f"'),
        # carriage_return == "\r"
        ('NAME', 'carriage_return'),
        ('DOUBLE_EQ', '=='),
        ('STRING_LITERAL', r'"\r"'),
        ('EQ_LINE', '===='),
        ]
    assert tv == tv_, tv


def test_boolean_div_setminus():
    lexer = _lex.Lexer()
    source = r'''
        /\ a / b
        /\ \/ c \ d
        '''
    tv = _tokenize(source, lexer)
    tv_ = [
        # /\ a / b
        ('AND', '/\\'),
        ('NAME', 'a'),
        ('SLASH', '/'),
        ('NAME', 'b'),
        # /\ \/ c \ d
        ('AND', '/\\'),
        ('OR', '\\/'),
        ('NAME', 'c'),
        ('BSLASH', '\\'),
        ('NAME', 'd'),
        ]
    assert tv == tv_, tv


def test_operator_double_at():
    lexer = _lex.Lexer()
    # `@@`
    text = r'''
        f @@ g
        '''
    tv = _tokenize(text, lexer)
    tv_ = [
        ('NAME', 'f'),
        ('DOUBLE_AT', '@@'),
        ('NAME', 'g'),
        ]
    assert tv == tv_, tv
    # `@@@`
    text = r'''
        f @@@ g
        '''
    tv = _tokenize(text, lexer)
    tv_ = [
        ('NAME', 'f'),
        ('DOUBLE_AT', '@@'),
        ('AT', '@'),
        ('NAME', 'g'),
        ]
    assert tv == tv_, tv


def _tokenize(
        source:
            str,
        lexer
        ) -> list[
            tuple[str, str]]:
    """Return `list` of pairs `(type, value)`.

    @type lexer:
        the method `lexer.parse()`
        returns an iterator
    """
    return [
        (token.symbol, token.value)
        for token in
            lexer.parse(source)]


if __name__ == '__main__':
    # test_lexing_module_example()
    # test_lexing_nested_multiline_comments()
    # test_lexing_nested_uniline_comments()
    # test_uniline_inside_multiline_comments()
    # test_multiline_inside_uniline_comments()
    # test_string_meta_sequences()
    # test_boolean_div_setminus()
    test_operator_double_at()



"""Miscellanea."""
import collections.abc as _abc
import functools as _ft
import inspect
import itertools as _itr
import logging as _log
import os
import pprint as _pp
import re
import typing as _ty


logger = _log.getLogger(__name__)


class Token(
        _ty.NamedTuple):
    symbol: str
    value: str
    row: int | None=None
    column: int | None=None
    start: int | None=None


def make_operator_precedence_grammar(
        precedence_associativity:
            tuple[str],
        tokens_of_symbols:
            dict[str, str],
        instance:
            _ty.Any
        ) -> None:
    """Add grammar equalities to `instance`.

    The new parser methods construct
    syntax-tree nodes by instantiating
    as follows:

    ```python
    instance._nodes.OperatorApplication(
        operator=...,
        arguments=...)
    ```

    @param precedence_associativity:
        names, used for:
        - lowercased as grammar symbols
        - uppercased as token names,
          unless lowercased they are
          keys of `tokens_of_symbols`

        The last item is not mapped to
        a new grammar equality.
    @param tokens_of_symbols:
        maps grammar symbols to token names
    @param tokens_of_symbols:
        with lowercase keys
        and uppercase values
    @param instance:
        parser to which to attach the
        new grammar equalities
    """
    pairs = _itr.pairwise(
        precedence_associativity)
    for operator, operand in pairs:
        operator = operator.lower()
        operand = operand.lower()
        _make_grammar_equality_for_op(
            operator, operand,
            tokens_of_symbols, instance)


def _make_grammar_equality_for_op(
        operator:
            str,
        operand:
            str,
        tokens_of_symbols:
            dict[str, str],
        instance:
            _ty.Any
        ) -> None:
    """Attach new grammar methods for operator.

    The newly created methods describe
    the grammar of `operator`, and are
    attached as bound methods to `instance`.

    @param operator:
        name of operator (lowercase)
    @param operand:
        name of operand (lowercase)

    Read the docstring of the function
    `make_operator_precedence_grammar`.
    """
    operator_token = _make_op_token(
        operator, tokens_of_symbols)
    _make_op = _pick_make_func(operator)
    functions = _make_op(
        operator, operand,
        operator_token)
    for function in functions:
        name = function.__name__
        if not _is_grammar_method_name(name):
            raise ValueError(
                name, function)
        # line numbers remain unchanged here
        _attach_as_method(function, instance)


def _pick_make_func(
        operator:
            str
        ) -> _abc.Callable:
    """Map `operator` to function."""
    *_, fixity = operator.rpartition('_')
    if fixity not in _FIXITY_TO_MK:
        raise ValueError(operator)
    return _FIXITY_TO_MK[fixity]


def _attach_as_method(
        function:
            _abc.Callable,
        instance:
            _ty.Any,
        attribute_name:
            str |
            None=None
        ) -> None:
    """Add `function` as method to `instance`.

    The new method is created by
    binding `function` to `instance`.

    @param attribute_name:
        name of the attribute of
        `instance` that will be assigned
        the new method
    """
    if attribute_name is None:
        attribute_name = function.__name__
    bound_func = function.__get__(instance)
    setattr(
        instance, attribute_name,
        bound_func)


def _make_prefix_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for prefix operator.

    This operator does not nest.

    ```
    G.operator_name =
        | L.OPERATOR_TOKEN & G.operand
            (* apply *)
        | G.operand
            (* skip *)
    ```

    @param operator_name:
        (lowercase) name of
        the grammar symbol of
        this production
    @param operand:
        (lowercase) name of
        the grammar symbol that
        is the production with
        precedence higher than
        this production
    @param operator_token:
        token of operator
    @return:
        unbound functions named as
        the intended parser methods
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    prefix_method = _make_fixity_method(
        operator_name, 'prefix')
    prefix_method.__doc__ = f'''
        {operator_name}:
            {operator_token}
            {operand}
        '''
    return [prefix_method, skip_method]


def _make_before_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for nestable prefix operator.

    This operator can nest.

    ```
    G.operator_name =
        | L.OPERATOR_TOKEN & G.operator_name
            (* apply; maybe nested *)
        | G.operand
            (* skip or start applying *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    prefix_method = _make_fixity_method(
        operator_name, 'prefix')
    prefix_method.__doc__ = f'''
        {operator_name}:
            {operator_token}
            {operator_name}
        '''
    return [prefix_method, skip_method]


def _make_left_assoc_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for left associative operator.

    This operator can nest.

    ```
    G.operator_name =
        | G.operator_name &
            L.OPERATOR_TOKEN &
            G.operand
                (* apply; maybe nested *)
        | G.operand
            (* skip or start applying *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    infix_method = _make_fixity_method(
        operator_name, 'infix')
    infix_method.__doc__ = f'''
        {operator_name}:
            {operator_name}
            {operator_token}
            {operand}
        '''
    return [infix_method, skip_method]


def _make_right_assoc_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for right associative operator.

    This operator can nest.

    ```
    G.operator_name =
        | G.operand &
            L.OPERATOR_TOKEN &
            G.operator_name
                (* apply; maybe nested *)
        | G.operand
            (* skip or start applying *)
    ```
    """
    # NOTE: for an LR(1) parser a more
    # efficient approach is to use left recursion,
    # and an intermediary production that
    # applies `reversed()` on the accumulated `list`,
    # and then folds the list into a tree.
    skip_method = _make_skip_method(
        operator_name, operand)
    infix_method = _make_fixity_method(
        operator_name, 'infix')
    infix_method.__doc__ = f'''
        {operator_name}:
            {operand}
            {operator_token}
            {operator_name}
        '''
    return [infix_method, skip_method]


def _make_nonassoc_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for nonassociative operator.

    This operator does not nest.

    ```
    G.operator_name =
        | G.operand &
            L.OPERATOR_TOKEN &
            G.operand
                (* apply *)
        | G.operand
            (* skip *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    infix_method = _make_fixity_method(
        operator_name, 'infix')
    infix_method.__doc__ = f'''
        {operator_name}:
            {operand}
            {operator_token}
            {operand}
        '''
    return [infix_method, skip_method]


def _make_postfix_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for postfix operator.

    This operator does not nest.

    ```
    G.operator_name =
        | G.operand & L.OPERATOR_TOKEN
            (* apply *)
        | G.operand
            (* skip *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    postfix_method = _make_fixity_method(
        operator_name, 'postfix')
    postfix_method.__doc__ = f'''
        {operator_name}:
            {operand}
            {operator_token}
        '''
    return [postfix_method, skip_method]


def _make_after_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    """Production for nestable postfix operator.

    This operator can nest.

    ```
    G.operator_name =
        | G.operator_name & L.OPERATOR_TOKEN
            (* apply; maybe nested *)
        | G.operand
            (* skip or start applying *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    postfix_method = _make_fixity_method(
        operator_name, 'postfix')
    postfix_method.__doc__ = f'''
        {operator_name}:
            {operator_name}
            {operator_token}
        '''
    return [postfix_method, skip_method]


def _make_infix_items_op(
        operator_name:
            str,
        operand:
            str,
        operator_token:
            str
        ) -> list[
            _abc.Callable]:
    r"""Production for infix itemized operator.

    This operator can repeat,
    similarly lexically to nestable operators.

    ```
    /\ G.operator_name =
        | G.items
            (* apply listy operator *)
        | G.operand
            (* skip *)
    /\ G.items =
        | G.items & L.OPERATOR_TOKEN & G.operand
            (* continue applying *)
        | G.operand & L.OPERATOR_TOKEN & G.operand
            (* start applying *)
    ```
    """
    skip_method = _make_skip_method(
        operator_name, operand)
    itemize, repeat, start = (
        _make_infix_items_method(
            operator_name))
    items = f'{operator_name}_items'
    itemize.__doc__ = f'''
        {operator_name}:
            {items}
        '''
    repeat.__doc__ = f'''
        {items}:
            {items}
            {operator_token}
            {operand}
        '''
    start.__doc__ = f'''
        {items}:
            {operand}
            {operator_token}
            {operand}
        '''
    return [
        itemize, repeat, start,
        skip_method]


_FIXITY_TO_MK: _ty.Final = dict(
    prefix=_make_prefix_op,
    before=_make_before_op,
    infix=_make_nonassoc_op,
    left=_make_left_assoc_op,
    between=_make_infix_items_op,
    right=_make_right_assoc_op,
    postfix=_make_postfix_op,
    after=_make_after_op,
    )


def _make_op_token(
        operator_name:
            str,
        tokens_of_symbols:
            dict[str, str]
        ) -> str:
    """Return name of token of `operator_name`.

    ```
    ASSUME
        operator_name.islower()
    ```
    """
    return tokens_of_symbols.get(
        operator_name, operator_name).upper()


def _make_fixity_method(
        operator_name:
            str,
        fixity:
            str
        ) -> _abc.Callable:
    r"""Return function for parser equation.

    ```
    ASSUME
        /\ operator_name.islower()
        /\ fixity.islower()
    ```

    @return:
        unbound function named as
        the intended parser method
    """
    match fixity:
        case 'prefix':
            _method = _prefix_method
        case 'infix':
            _method = _infix_method
        case 'postfix':
            _method = _postfix_method
        case _:
            raise ValueError(fixity)
    def method(self, p):
        _method(self, p)
    method.__name__ = f'p_{operator_name}'
    return method


def _make_infix_items_method(
        operator_name:
            str
        ) -> list:
    """Return functions for parser equation.

    ```
    ASSUME
        operator_name.islower()
    ```

    @return:
        `list` of unbound functions,
        named as the intended parser methods
    """
    def itemize(self, p):
        # docstring dynamically assigned to wrapper
        # grammar equality of the form
        # nonleaf: nonleaf_repetition
        op_lexeme, *args = p[1]
        p[0] = self._nodes.OperatorApplication(
            operator=op_lexeme,
            arguments=args)
    def append(self, p):
        _repetition_append_method(self, p)
    def start(self, p):
        _repetition_start_method(self, p)
    itemize.__name__ = f'p_{operator_name}'
    append.__name__ = f'p_{operator_name}_append'
    start.__name__ = f'p_{operator_name}_start'
    return [itemize, append, start]


def _prefix_method(
        self, p
        ) -> None:
    # docstring dynamically assigned to wrapper
    # grammar equality of the form
    # nonleaf: OPERATOR operand
    p[0] = self._nodes.OperatorApplication(
        operator=p[1],
        arguments=[p[2]])


def _infix_method(
        self, p
        ) -> None:
    # docstring dynamically assigned to wrapper
    # grammar equality of the form
    # nonleaf: operand OPERATOR operand
    p[0] = self._nodes.OperatorApplication(
        operator=p[2],
        arguments=[p[1], p[3]])


def _postfix_method(
        self, p
        ) -> None:
    # docstring dynamically assigned to wrapper
    # grammar equality of the form
    # nonleaf: operand OPERATOR
    p[0] = self._nodes.OperatorApplication(
        operator=p[2],
        arguments=[p[1]])


def _repetition_append_method(
        self, p
        ) -> None:
    # docstring dynamically assigned to wrapper
    # grammar equality of the form
    # nonleaf_repetition:
    #     nonleaf_repetition OPERATOR operand
    p[1].append(p[3])
    p[0] = p[1]


def _repetition_start_method(
        self, p
        ) -> None:
    # docstring dynamically assigned to wrapper
    # grammar equality of the form
    # nonleaf_repetition: operand
    p[0] = [p[2], p[1], p[3]]


def _make_skip_method(
        operator_name:
            str,
        operand:
            str
        ) -> _abc.Callable:
    r"""Production that skips to `operand`.

    ```
    ASSUME
        /\ operator_name.islower()
        /\ operand.islower()
    ```

    @return:
        unbound function
    """
    def skip_method(self, p):
        p[0] = p[1]
    method_name = f'p_{operator_name}_skip'
    method_docstring = f'''
        {operator_name}: {operand}
        '''
    skip_method.__name__ = method_name
    skip_method.__doc__ = method_docstring
    return skip_method


def format_bnf_docstrings(
        grammar
        ) -> None:
    """Infix pipe, splice lines, space around `:`.

    @param grammar:
        instance, to format the docstrings of
        `p_*` methods
    """
    for name, method in _methods_of(grammar):
        if not _is_grammar_method_name(name):
            continue
        doc = method.__doc__
        if doc is None:
            continue
        doc = _transform_bnf(doc)
        method.__func__.__doc__ = doc


def _transform_bnf(
        doc:
            str
        ) -> str:
    """Infix pipe, splice lines, space ` : `."""
    doc = re.sub(
        r' ([^\s]) : ',
        r'\1 :',
        doc,
        flags=re.VERBOSE)
    doc = re.sub(
        r' : \s* \| ',
        ': ',
        doc,
        flags=re.VERBOSE)
    return re.sub(
        r' \n \s* ([^|\s]) ',
        r' \1',
        doc,
        flags=re.VERBOSE)


def check_bnf_grammar(
        grammar
        ) -> None:
    """Assert no `grammar._p_*` methods.

    Raises `AssertionError` if any methods
    are found whose names start with `_p_`.
    """
    for name, method in _methods_of(grammar):
        if not name.startswith('_p_'):
            continue
        raise AssertionError(name, method)


def make_token_grammar(
        lexemes_to_tokens:
            dict,
        tokens_to_regexes:
            dict,
        instance
        ) -> None:
    """Add tokenization methods for literals.

    @param instance:
        grammar class
    """
    lextok = lexemes_to_tokens
    regexes = tokens_to_regexes
    for token_type, regex in regexes.items():
        if regex is None:
            continue
        def method(self, token):
            return token
        name = _name_token_var(token_type)
        method.__name__ = name
        method.__doc__ = regex
        bound_func = method.__get__(instance)
        setattr(instance, name, bound_func)
    lexemes = sort_lexemes(lextok)
    for lexeme in lexemes:
        token_type = lextok[lexeme]
        name = _name_token_var(token_type)
        regex = re.escape(lexeme)
        # NOTE: `instance.__dict__` records
        # insertion order by `setattr()`
        if hasattr(instance, name):
            raise ValueError(name, regex)
        setattr(instance, name, regex)


def _name_token_var(
        token_type
        ) -> str:
    """Return lexer method name."""
    return f't_{token_type}'


def sort_lexemes(
        lexemes:
            _abc.Iterable
        ) -> list:
    """Sort `lexemes` in descending length."""
    return sorted(
        lexemes,
        key=len,
        reverse=True)


def _format_docstring(
        *,
        _=None,
        **kw
        ) -> _abc.Callable:
    """Format doctsring using `**kw`.

    Formats the docstring of function
    `func`, using the key-value pairs
    in the dictionary `kw` as keyword
    parameter-argument pairs passed
    to the method `func.__doc__.format`.

    This decorator is needed because
    formatted string literals cannot
    be docstrings.
    """
    if _ is not None:
        raise TypeError(
            'only keyword arguments '
            'other than `_`')
    def _decorator(func):
        if func.__doc__ is not None:
            func.__doc__ = (
                func.__doc__.format(**kw))
            return func
        raise ValueError(
            f'`{func}` appears to '
            'not have a docstring')
    return _decorator


def dumps_grammar(
        grammar,
        add_braces:
            bool=False
        ) -> str:
    """Return equalities of `grammar`.

    @param grammar:
        instance with `p_*` methods
    """
    (tokens,
        start_symbol,
        nonleaf_symbols,
        grammar_text,
        ) = collect_grammar_info(
            grammar, add_braces)
    return grammar_text


def collect_grammar_info(
        grammar:
            _ty.Any,
        add_braces:
            bool
        ) -> tuple[
            set[str],
            str,
            set[str],
            str]:
    """Return grammar symbols and equalities.

    @param grammar:
        the information is collected from
        the `p_*` methods of this object
    @param add_braces:
        if `True`,
        then append `{ }` to each production
        (for semantic actions)
    @return:
        ```py
        (leaf_symbols,
         root_symbol,
         nonleaf_symbols,
         grammar_text)
        ```
    """
    root_symbol = grammar.start
    methods = list(_methods_of(grammar))
    # print(f'{len(methods)} grammar rules')
    def sieve(x):
        name, _ = x
        return _is_grammar_method_name(name)
    methods = filter(sieve, methods)
    def symbol_of(x):
        _, method = x
        docstring = method.__doc__
        if docstring.count(':') != 1:
            raise ValueError(docstring)
        symbol, _ = docstring.split(':')
        return symbol.strip()
    def row(x):
        _, method = x
        _, start_line = inspect.getsourcelines(
            method)
        return start_line
    methods = sorted(
        methods,
        key=row)
    # methods = list(methods)
    # for name, _ in methods:
    #     print(name)
    INDENT = ' ' * 4
    def format_line(line):
        line = line.strip()
        if '=' in line:
            line = f'/\\ {line}'
        if line.startswith('|'):
            line = f'{INDENT}{line}'
        return line
    grammar_symbols = set()
    def format_expansion(
            expansion:
                str
            ) -> str:
        symbols = expansion.split()
        grammar_symbols.update(symbols)
        return ' & '.join(symbols)
    nonleaf_symbols = set()
    out_lines = list()
    last_symbol = None
    for _, method in methods:
        docstring = method.__doc__
        if docstring.count(':') != 1:
            raise ValueError(docstring)
        symbol, expansion = docstring.split(':')
        symbol = symbol.strip()
        expansions = expansion.split('|')
        expansions = map(
            format_expansion,
            expansions)
        expansion = '\n| '.join(expansions)
        if symbol == last_symbol:
            docstring = f'| {expansion}'
        else:
            docstring = f'{symbol} = {expansion}'
        lines = docstring.splitlines()
        lines = map(format_line, lines)
        lines = filter(
            lambda x: x,
            lines)
        if add_braces:
            lines = map(
                lambda x: x + ' { }',
                lines)
        out_lines.extend(lines)
        nonleaf_symbols.add(symbol)
        last_symbol = symbol
    leaf_symbols = grammar_symbols - nonleaf_symbols
    grammar_text = '\n'.join(out_lines)
    return (
        leaf_symbols,
        root_symbol,
        nonleaf_symbols,
        grammar_text)


def _is_grammar_method_name(
        name:
            str
        ) -> bool:
    """Return `True` if a grammar method name."""
    return (
        name.startswith('p_') and
        name != 'p_error')


def _methods_of(
        instance
        ) -> _abc.Iterable:
    """Return iterator over methods."""
    yield from inspect.getmembers(
        instance,
        predicate=inspect.ismethod)


def rewrite_tables(
        parser_class,
        ) -> None:
    """Write parser table to file.

    Overwrite pre-existing file.
    """
    parser = parser_class()
    filepath = parser.cache_filepath
    if os.path.isfile(filepath):
        logger.info(
            f'found file `{filepath}`')
        os.remove(filepath)
        logger.info(
            f'removed file `{filepath}`')
    parser.build()
    if not os.path.isfile(filepath):
        raise AssertionError(
            'rebuilt parser cache not '
            f'found at: {filepath}')


def configure_logging(
        logger_name:
            str
        ) -> None:
    """Setup logging for `logger_name`."""
    logger = _log.getLogger(logger_name)
    logger.setLevel(_log.INFO)
    handler = _log.StreamHandler()
    fmt = '[%(levelname)s]: %(message)s'
    formatter = _log.Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def add_column_to_tokens(
        tokens:
            _abc.Iterable[
                Token],
        string:
            str
        ) -> _abc.Iterable:
    """Generate tokens annotated with columns.

    Returns an iterator of new token objects
    that have also an attribute `token.column`.
    """
    string_length = len(string)
    row = 0
    line_start = 0
    newline_indices = _index_newlines(string)
    for token in tokens:
        if row > token.row:
            raise ValueError(
                token.row, row)
        while row < token.row:
            row += 1
            index = next(newline_indices)
            line_start = index + 1
        column = (token.start - line_start) + 1
        if column < 0 or column >= string_length:
            raise AssertionError(
                'expected '
                rf'column \in 1..{string_length} '
                f'(computed `{column = }`)')
        yield copy_lex_token(
            token,
            column=column)


def _index_newlines(
        string:
            str
        ) -> _abc.Iterable:
    r"""Generate indices of newlines in `string`.

    Returns an iterator of indices of the
    newline characters `\n` that are in `string`.
    """
    for m in re.finditer('\n', string):
        yield m.start()


def find_token_column(
        string:
            str,
        token:
            Token
        ) -> int:
    r"""Return start column of `token`.

    @param string:
        input to lexer
    @return:
        number of column where
        `token` starts, with
        `column \in 1..len(string)`
    """
    index = string.rfind('\n', 0, token.start)
    line_start = index + 1
    column = (token.start - line_start) + 1
    string_length = len(string)
    if column < 1 or column > string_length:
        raise AssertionError(
            'expected'
            rf'`column \in 1..{string_length}` '
            f'(computed `{column = }`)')
    return column


def copy_lex_token(
        token:
            Token,
        **kw
        ) -> Token:
    """Return a copy of `token`.

    The new token has a `column` attribute.
    Keyword arguments in `kw` change the
    values of attributes of `token`.
    """
    column = getattr(token, 'column', None)
    attrs = dict(
        symbol=token.symbol,
        value=token.value,
        row=token.row,
        column=column,
        start=token.start)
    attrs.update(kw)
    new_token = Token(**attrs)
    return new_token


def make_lex_token(
        token_type:
            str,
        token_value:
            str |
            None=None,
        row:
            int=-1,
        start:
            int=-1,
        column:
            int |
            None=None
        ) -> Token:
    """Return a modified token.

    If `token_value is None`, then the
    returned token has `token_type` as
    valued.
    """
    if token_value is None:
        token_value = token_type
    return Token(
        symbol=token_type,
        value=token_value,
        row=row,
        column=column,
        start=start)


def pipe(
        x,
        *pipeline:
            _abc.Callable):
    """Apply each item of `pipeline` to `x`.

    Example:

    ```python
    import operator

    x = 1
    pipeline = [operator.neg, hex]
    y = pipe(x, *pipeline)
    assert y == '-0x1', y
    ```

    @param x:
        value for input to `pipeline[0]`
    """
    return _ft.reduce(
        _postfix_apply, pipeline, x)


def _postfix_apply(
        x,
        f:
            _abc.Callable):
    """Return `f(x)`."""
    return f(x)


def reverse(
        iterable:
            _abc.Iterable
        ) -> list:
    """Return `reversed(list(iterable))`."""
    return reversed(list(iterable))


def printer(
        items:
            _abc.Iterable
        ) -> list:
    """Print and return `list(items)`."""
    items = list(items)
    start_line = 10 * '-'
    print(start_line)
    for item in items:
        print(item)
    end_line = 10 * '='
    print(end_line)
    return items


def map_dict(
        key_mapper:
            _abc.Callable,
        value_mapper:
            _abc.Callable,
        kv:
            dict
        ) -> dict:
    """Map keys and values of `kv`."""
    return {
        key_mapper(k): value_mapper(v)
        for k, v in kv.items()}


def map_keys(
        key_mapper:
            _abc.Callable,
        kv:
            dict
        ) -> dict:
    """Map keys of `kv` (same values)."""
    return {
        key_mapper(k): v
        for k, v in kv.items()}


def map_values(
        value_mapper:
            _abc.Callable,
        kv:
            dict
        ) -> dict:
    """Map values of `kv` (same keys)."""
    return {
        k: value_mapper(v)
        for k, v in kv.items()}


def assert_len(
        container:
            _abc.Container,
        expected_length:
            int
        ) -> None:
    """Assert `container` has `expected_length`.

    Raises `AssertionError` if
    `len(container) != expected_length`.

    Example:

    ```python
    import tla._utils as _utils


    _MODULE_CONSTANT = {
        'item_1',
        'item_2',
        'iten_3',  # typo
        }
    _utils.assert_len(_MODULE_CONSTANT, 3)
    _MORE = _MODULE_CONSTANT | {
        'item_4',
        'item_5',
        }
    _utils.assert_len(
        _MORE,
        2 + len(_MODULE_CONSTANT))
    ```

    ```
    ASSUME
        expected_length >= 0
    ```
    """
    length = len(container)
    if length == expected_length:
        return
    formatted = _pp.pformat(container)
    message = (
        'container has unexpected length:\n'
        f'expected length:  {expected_length}\n'
        f'actual length:  {length}\n'
        f'Typo ?\n'
        f'The container is:\n{formatted}')
    raise AssertionError(message)


def rindex(
        value,
        sequence
        ) -> int:
    """Return first index, searching from end."""
    n = len(sequence)
    for index in range(n - 1, -1, -1):
        if sequence[index] == value:
            return index
    raise ValueError(
        f'{value} is not in sequence')



"""Test `tla._preparser`."""
import parstools._lex as _lex
import tla._preparser as _prp


Token = _lex.Token


def test_insert_def_sig_tokens_nullary():
    in_tokens = [
        Token('OP_15_POSTFIX', '^+'),
        Token('NAME', 'a'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_fixal(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(1, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens


def test_insert_def_sig_tokens_prefix():
    in_tokens = [
        Token('OP_12_PREFIX', '-.'),
        Token('NAME', 'a'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_fixal(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens


def test_insert_def_sig_tokens_infix():
    in_tokens = [
        Token('NAME', 'a'),
        Token('OP_10_LEFT', '+'),
        Token('NAME', 'b'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_fixal(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens


def test_insert_def_sig_tokens_postfix():
    in_tokens = [
        Token('NAME', 'a'),
        Token('OP_15_POSTFIX', '^*'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_fixal(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens


def test_insert_def_sig_tokens_nonfix():
    # one parameter
    in_tokens = [
        Token('NAME', 'Op'),
        Token('LPAREN', '('),
        Token('NAME', 'x'),
        Token('RPAREN', ')'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_nonfix(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens
    # multiple parameters
    in_tokens = [
        Token('NAME', 'Op'),
        Token('LPAREN', '('),
        Token('NAME', 'x'),
        Token('COMMA', ','),
        Token('NAME', 'y'),
        Token('RPAREN', ')'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_nonfix(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens
    # second-order operator
    # (so nested parentheses)
    in_tokens = [
        Token('NAME', 'Op'),
        Token('LPAREN', '('),
        Token('NAME', 'g'),
        Token('LPAREN', '('),
        Token('NAME', 'x'),
        Token('RPAREN', ')'),
        Token('COMMA', ','),
        Token('NAME', 'y'),
        Token('RPAREN', ')'),
        Token('DOUBLE_EQ', '==')]
    tokens = _prp._insert_def_sig_tokens_nonfix(
        reversed(in_tokens))
    tokens = list(reversed(list(tokens)))
    tokens_ = list(in_tokens)
    tokens_.insert(0, Token('DEF_SIG', 'DEF_SIG'))
    assert tokens == tokens_, tokens


def test_swap_local_keyword():
    in_tokens = [
        Token('LOCAL', 'LOCAL'),
        Token('DEF_SIG', 'DEF_SIG')]
    tokens = _prp.swap_local_keyword(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('DEF_SIG', 'DEF_SIG'),
        Token('LOCAL', 'LOCAL')]
    assert tokens == tokens_, tokens


def test_insert_proof_define_tokens():
    in_tokens = [
        Token('STEP_NUMBER', '<1>1.'),
        Token('DEF_SIG', 'DEF_SIG'),
        ]
    tokens = _prp.insert_proof_define_tokens(in_tokens)
    tokens = [
        token.symbol
        for token in tokens]
    tokens_ = [
        'STEP_NUMBER',
        'DEFINE',  # inserted token
        'DEF_SIG']
    assert tokens == tokens_, tokens


def test_end_define_proof_steps():
    in_tokens = [
        Token('DEFINE', 'DEFINE'),
        Token('NAME', 'a'),
        Token('DOUBLE_EQ', '=='),
        Token('FLOAT', '1.0'),
        Token('STEP_NUMBER', '<1>1.'),
        ]
    tokens = _prp.end_define_proof_steps(in_tokens)
    tokens = list(tokens)
    tokens = [
        token.symbol
        for token in tokens]
    tokens_ = [
        'DEFINE',
        'NAME',
        'DOUBLE_EQ',
        'FLOAT',
        'END_DEFINE',  # inserted token
        'STEP_NUMBER']
    assert tokens == tokens_, tokens


def test_switch_step_identifiers():
    in_tokens = [
        Token('BY', 'BY'),
        Token('NAME', 'theorem_name'),
        Token('COMMA', ','),
        Token('STEP_NUMBER', '<1>1'),
        ]
    tokens = _prp.switch_step_identifiers(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('BY', 'BY'),
        Token('NAME', 'theorem_name'),
        Token('COMMA', ','),
        Token('NAME', '<1>1'),  # changed
        ]
    assert tokens == tokens_, tokens


def test_join_as_one_token():
    # `! @`
    in_tokens = [
        Token('EXCLAMATION', '!'),
        Token('AT', '@'),
        ]
    tokens = _prp.join_as_one_token(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('EXCLAMATION_AT', '@'),
        ]
    assert tokens == tokens_, tokens
    # `! :`
    in_tokens = [
        Token('EXCLAMATION', '!'),
        Token('COLON', ':'),
        ]
    tokens = _prp.join_as_one_token(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('EXCLAMATION_COLON', ':'),
        ]
    assert tokens == tokens_, tokens
    # `! <<`
    in_tokens = [
        Token('EXCLAMATION', '!'),
        Token('DOUBLE_LANGLE', '<<'),
        ]
    tokens = _prp.join_as_one_token(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('EXCLAMATION_DOUBLE_LANGLE', '<<'),
        ]
    assert tokens == tokens_, tokens
    # `! >>`
    in_tokens = [
        Token('EXCLAMATION', '!'),
        Token('DOUBLE_RANGLE', '>>'),
        ]
    tokens = _prp.join_as_one_token(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('EXCLAMATION_DOUBLE_RANGLE', '>>'),
        ]
    assert tokens == tokens_, tokens
    # `! NAME`
    in_tokens = [
        Token('EXCLAMATION', '!'),
        Token('NAME', 'a'),
        ]
    tokens = _prp.join_as_one_token(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('EXCLAMATION_NAME', 'a'),
        ]
    assert tokens == tokens_, tokens


def test_switch_fixal_operators():
    in_tokens = [
        Token('NAME', 'Op'),
        Token('LPAREN', '('),
        Token('OP_10_LEFT', '+'),
        Token('COMMA', ','),
        Token('INTEGRAL_NUMERAL', '1'),
        Token('COMMA', ','),
        Token('INTEGRAL_NUMERAL', '2'),
        Token('RPAREN', ')'),
        ]
    tokens = _prp.switch_fixal_operators(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('NAME', 'Op'),
        Token('LPAREN', '('),
        Token('NAME', '+'),  # changed
        Token('COMMA', ','),
        Token('INTEGRAL_NUMERAL', '1'),
        Token('COMMA', ','),
        Token('INTEGRAL_NUMERAL', '2'),
        Token('RPAREN', ')'),
        ]
    assert tokens == tokens_, tokens


def test_switch_keywords_after_dot():
    in_tokens = [
        Token('DOT', '.'),
        Token('LET', 'LET'),
        ]
    tokens = _prp.switch_keywords_after_dot(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('DOT', '.'),
        Token('NAME', 'LET'),  # changed
        ]
    assert tokens == tokens_, tokens


def test_filter_comments():
    in_tokens = [
        Token('UNILINE_COMMENT', r'\* comment\n'),
        Token('MULTILINE_COMMENT', '(* comment *)'),
        Token('LET', 'LET'),
        ]
    tokens = _prp.filter_comments(in_tokens)
    tokens = list(tokens)
    tokens_ = [
        Token('LET', 'LET'),
        ]
    assert tokens == tokens_, tokens


if __name__ == '__main__':
    # test_filter_comments()
    test_insert_def_sig_tokens_infix()



"""How to parse a TLA+ module."""
import tla


TLA_FILE_PATH = 'Counter.tla'


def parse_module():
    """Parse a TLA+ module."""
    tla_text = _load_tla_module()
    tree = tla.parse(tla_text)
    print(tree)


def parse_module_and_format():
    """Parse, format, print a TLA+ module."""
    tla_text = _load_tla_module()
    tree = tla.parse(tla_text)
    text = tla.pformat_ast(tree)
    print(text)


def _load_tla_module():
    """Return contents of TLA+ file."""
    filepath = TLA_FILE_PATH
    with open(filepath, 'r') as fd:
        return fd.read()


if __name__ == '__main__':
    parse_module()
    parse_module_and_format()



"""Lexer for TLA+."""
import argparse as _arg
import logging
import time

import parstools._lex as _lex
import tla._langdef as _langdef
import tla._utils as _utils


logger = logging.getLogger(__name__)


COMMENT_TOKENS = {
    'UNILINE_COMMENT':
        r' \\ \* [^\n]* \n',
    'MULTILINE_COMMENT':
        None,
        # `(* ... *)`
    }
_utils.assert_len(COMMENT_TOKENS, 2)
_LITERAL_LEXEMES = {
    'BINARY_INTEGER':
        r'''
        (
              \\b
            | \\B
        )
        [0-1]+
        ''',
    'DASH_LINE':
        r' \-\-\-\-(\-)* ',
    'EQ_LINE':
        r' ====(=)* ',
    'HEXADECIMAL_INTEGER':
        r'''
        (
              \\h
            | \\H
        )
        (
              [0-9]
            | [A-F]
        )+
        ''',
    'OCTAL_INTEGER':
        r'''
        (
              \\o
            | \\O
        )
        [0-7]+
        ''',
    'STEP_NUMBER':
        rf'''
        <[0-9]+>
        [a-zA-Z0-9_]*
        \.*
        ''',
    'STEP_NUMBER_ASTERISK':
        r'''
        <\*>
        \.*
        ''',
    'STEP_NUMBER_PLUS':
        r'''
        <\+>
        \.*
        ''',
    }
_utils.assert_len(_LITERAL_LEXEMES, 8)
_OTHER_TOKENS = {
    'FLOAT',
    'DECIMAL_INTEGER',
    'NAME',
    'MULTILINE_STRING_LITERAL',
    'SF_',
    'STRING_LITERAL',
    'WF_',
    }
_BSLASH_SUFFIXES = {
    'A',
    'AA',
    'approx',
    'asymp',
    'bigcirc',
    'bullet',
    'cap',
    'cdot',
    'circ',
        # synonym to `\o`
    'cong',
    'cup',
    'div',
    'doteq',
    'E',
    'EE',
    'equiv',
        # synonym to `<=>`
    'geq',
        # synonym to `>=`
    'gg',
    'in',
    'intersect',
        # synonym to `\cap`
    'land',
        # synonym to `/\`
    'leadsto',
        # synonym to `~>`
    'leq',
        # synonym to `<=`
    'll',
    'lnot',
        # synonym to `~`
    'lor',
        # synonym to `\/`
    'mod',
        # synonym to `%`
    'neg',
        # synonym to `~`
    'notin',
    'o',
    'odot',
        # synonym to `(.)`
    'ominus',
        # synonym to `(-)`
    'oplus',
        # synonym to `(+)`
    'oslash',
        # synonym to `(\)`
    'otimes',
        # synonym to `(\X)`
    'prec',
    'preceq',
    'propto',
    'setminus',
        # synonym to `\`
    'sim',
    'simeq',
    'sqcap',
    'sqcup',
    'sqsubset',
    'sqsubseteq',
    'sqsupset',
    'sqsupseteq',
    'star',
    'subset',
    'subseteq',
    'succ',
    'succeq',
    'supset',
    'supseteq',
    'times',
        # synonym to `\X`
    'union',
        # synonym to `\cup`
    'uplus',
    'wr',
    'X',
        # CARTESIAN
    }
_utils.assert_len(_BSLASH_SUFFIXES, 59)
_PICTORIAL_LEXEMES = {
    '=>':
        'EQ_RANGLE',
    '<=>':
        'LANGLE_EQ_RANGLE',
    '~':
        'TILDE',
    "'":
        'PRIME',
    '^+':
        'CIRCUMFLEX_PLUS',
    '^*':
        'CIRCUMFLEX_ASTERISK',
    '^#':
        'CIRCUMFLEX_HASHMARK',
    '(+)':
        'OPLUS',
    '(-)':
        'ODASH',
    '(/)':
        'OSLASH',
    '(.)':
        'ODOT',
    r'(\X)':
        'OCARTESIAN',
    '/':
        'SLASH',
    '//':
        'DOUBLE_SLASH',
    '\\':
        'BSLASH',
    '/\\':
        'AND',
    r'\/':
        'OR',
    '??':
        'DOUBLE_QUESTION',
    '.':
        'DOT',
    '..':
        'DOUBLE_DOT',
    '...':
        'TRIPLE_DOT',
    '=':
        'EQ',
    '==':
        'DOUBLE_EQ',
    '|':
        'PIPE',
    '||':
        'DOUBLE_PIPE',
    '&':
        'ET',
    '&&':
        'DOUBLE_ET',
    '$':
        'DOLLAR',
    '$$':
        'DOUBLE_DOLLAR',
    '@':
        'AT',
    '@@':
        'DOUBLE_AT',
    '+':
        'PLUS',
    '++':
        'DOUBLE_PLUS',
    '-.':
        'DASH_DOT',
    '-':
        'DASH',
    '--':
        'DOUBLE_DASH',
    '*':
        'ASTERISK',
    '**':
        'DOUBLE_ASTERISK',
    '^':
        'CIRCUMFLEX',
    '^^':
        'DOUBLE_CIRCUMFLEX',
    '!':
        'EXCLAMATION',
    '!!':
        'DOUBLE_EXCLAMATION',
    '%':
        'PERCENT',
    '%%':
        'DOUBLE_PERCENT',
    '/=':
        'NEQ',
    '#':
        'HASHMARK',
    '##':
        'DOUBLE_HASHMARK',
    '|-':
        'PIPE_DASH',
    '|=':
        'PIPE_EQ',
    '-|':
        'DASH_PIPE',
    '=|':
        'EQ_PIPE',
    ':':
        'COLON',
    '::':
        'DOUBLE_COLON',
    ':=':
        'COLON_EQ',
    '::=':
        'DOUBLE_COLON_EQ',
    '<:':
        'LANGLE_COLON',
    ':>':
        'COLON_RANGLE',
    ',':
        'COMMA',
    '<-':
        'LARROW',
    '->':
        'RARROW',
    '|->':
        'MAPSTO',
    '-+->':
        'ARROW_PLUS',
    '~>':
        'LEADSTO',
    '(':
        'LPAREN',
    ')':
        'RPAREN',
    '{':
        'LBRACE',
    '}':
        'RBRACE',
    '[':
        'LBRACKET',
    ']':
        'RBRACKET',
    ']_':
        'RBRACKET_UNDERSCORE',
    '[]':
        'LBRACKET_RBRACKET',
    '<':
        'LANGLE',
    '>':
        'RANGLE',
    '<>':
        'LANGLE_RANGLE',
    '<<':
        'DOUBLE_LANGLE',
    '>>':
        'DOUBLE_RANGLE',
    '>>_':
        'DOUBLE_RANGLE_UNDERSCORE',
    '=<':
        'EQ_LANGLE',
    '<=':
        'LANGLE_EQ',
    '>=':
        'RANGLE_EQ',
    '_':
        'UNDERSCORE',
    }
_utils.assert_len(_PICTORIAL_LEXEMES, 81)
_SYNONYMS = {
    '<=>':
        r'\equiv',
    '/\\':
        r'\land',
    r'\/':
        r'\lor',
    # '#':
    #     '/=',
    '\\':
        r'\setminus',
    r'\cap':
        r'\intersect',
    r'\cup':
        r'\union',
    r'\X':
        r'\times',
    '~>':
        r'\leadsto',
    '>=':
        r'\geq',
    '%':
        r'\mod',
    '(+)':
        r'\oplus',
    '(-)':
        r'\ominus',
    '(.)':
        r'\odot',
    '(/)':
        r'\oslash',
    r'(\X)':
        r'\otimes',
    r'\circ':
        r'\o',
    '~':
        (r'\neg', r'\lnot'),
    '=<':
        (r'\leq',),  # '<='
    }
_utils.assert_len(_SYNONYMS, 18)
_TOKEN_RENAMING = {
    'BSLASH_CAP':
        'CAP',
    'BSLASH_CIRC':
        'CIRC',
    'BSLASH_CUP':
        'CUP',
    'BSLASH_TIMES':
        'CARTESIAN',
    'BSLASH_X':
        'CARTESIAN',
    # 'HASHMARK':
    #     'NEQ',
    }
_utils.assert_len(_TOKEN_RENAMING, 5)


def map_lexemes_to_tokens(
        ) -> dict[
            str, str]:
    """Return mapping lexemes to token names."""
    lextok = {
        rf'\{s}':
            f'BSLASH_{s.upper()}'
        for s in _BSLASH_SUFFIXES}
    lextok |= _PICTORIAL_LEXEMES
    op_synonyms = _make_synonymous_lexemes()
    for op, synonyms in op_synonyms.items():
        for other in synonyms:
            token_name = lextok[op]
            _add_to_dict(other, token_name, lextok)
    def mapper(value):
        return _TOKEN_RENAMING.get(value, value)
    lextok = _utils.map_values(mapper, lextok)
    return lextok


def _make_synonymous_lexemes(
        ) -> dict[
            str, str]:
    """Return `dict` of synonyms."""
    def mapper(x):
        if isinstance(x, str):
            return (x,)
        return x
    return _utils.map_values(
        mapper, _SYNONYMS)


def _add_to_dict(
        key,
        value,
        mapping):
    """Set `mapping[key] = value`.

    Raise `AssertionError` if `key`
    is already in `mapping`.
    """
    add = (
        key not in mapping or
        _is_bslash_operator(key))
    if add:
        mapping[key] = value
        return
    raise AssertionError(
        f'{key = }, '
        f'{mapping = }')


# lexer states
_INITIAL_LEXER_STATE = 'initial'
_COMMENT_LEXER_STATE = 'multilinecomment'
_STRING_LEXER_STATE = 'string'
_MULTILINE_STRING_LEXER_STATE = 'multilinestring'


class Lexer(
        _lex.StatefulLexer):
    """Grammar to build TLA+ lexer."""

    lexing_states = (
        _INITIAL_LEXER_STATE,
        _COMMENT_LEXER_STATE,
        _STRING_LEXER_STATE,
        _MULTILINE_STRING_LEXER_STATE)

    def __init__(
            self
            ) -> None:
        self._lextok = map_lexemes_to_tokens()
        self.tokens = sorted(set().union(
            _langdef.KEYWORDS,
            COMMENT_TOKENS,
            _LITERAL_LEXEMES,
            _OTHER_TOKENS,
            self._lextok.values(),
            ))
        self.token_types = self.tokens
        self.states = (
            (_COMMENT_LEXER_STATE,
                'exclusive'),
            (_STRING_LEXER_STATE,
                'exclusive'),
            (_MULTILINE_STRING_LEXER_STATE,
                'exclusive'))
        self._keywords = set(_langdef.KEYWORDS)
        self._make_token_vars()
        # state
        self._comment_level: int = 0
        self._comment_start: int | None = None
        self._line_number: int = 0
        self._string_start: int | None = None
        super().__init__()

    def _make_token_vars(
            self
            ) -> None:
        regexes = dict(_LITERAL_LEXEMES)
        regexes.update(COMMENT_TOKENS)
        lextok = {
            k: v
            for k, v in self._lextok.items()
            if not _is_bslash_operator(k)}
        _utils.make_token_grammar(
            lextok, regexes, self)

    def parse(
            self,
            text:
                str):
        """Return tokens for `text`.

        To tokenize the lexeme `ELIF` as
        an identifier:

        ```python
        self._keywords.remove('ELIF')
        ```
        """
        # reset the lexer's state
        self._line_number = 0
        self._string_start = None
        tokens = super().parse(text)
        return _lex.add_row_column(
            text, tokens)

    def t_COMMENT(
            self,
            token):
        r' \( \* '
        self._comment_level = 1
        self.lexing_state = _COMMENT_LEXER_STATE
        self._comment_start = token.start

    def t_multilinecomment_left_paren(
            self,
            token):
        r' \( \* '
        self._comment_level += 1

    def t_multilinecomment_right_paren(
            self,
            token):
        r' \* \) '
        self._comment_level -= 1
        if self._comment_level < 0:
            raise AssertionError(
                self._comment_level)
        if self._comment_level > 0:
            return
        self.lexing_state = _INITIAL_LEXER_STATE
        # make token
        start = self._comment_start
        end = token.start + len(token.value)
        value = self.input_text[start:end]
        return _lex.Token(
            symbol='MULTILINE_COMMENT',
            value=value,
            start=start)

    def t_multilinecomment_newline(
            self,
            token):
        r' \n '
        self._line_number += 1

    def t_multilinecomment_other(
            self,
            token):
        r' [^\n(\*]+ '
        return None

    def t_multilinecomment_lparen(
            self,
            token):
        r' \( '
        return None

    def t_multilinecomment_asterisk(
            self,
            token):
        r' \* '
        return None

    def t_MULTILINE_STRING_LITERAL(
            self,
            token):
        r' """ '
        self._string_start = token.start
        self.lexing_state = _MULTILINE_STRING_LEXER_STATE

    def t_multilinestring_escaped_quotes(
            self,
            token):
        r"""
          \\ t
        | \\ n
        | \\ f
        | \\ r
        | \\ \\
        """

    def t_multilinestring_newline(
            self,
            token):
        r' \n '

    def t_multilinestring_other_characters(
            self,
            token):
        r' [^"\\]+ '

    def t_multilinestring_closing(
            self,
            token):
        r' """ '
        return self._collect_string(
            token, 'MULTILINE_STRING_LITERAL')

    def t_multilinestring_quotes(
            self,
            token):
        r"""
          ""
        | "
        """

    def t_STRING_LITERAL(
            self,
            token):
        r' " '
        self._string_start = token.start
        self.lexing_state = _STRING_LEXER_STATE

    def t_string_escaped_quotes(
            self,
            token):
        r"""
          \\ "
        | \\ t
        | \\ n
        | \\ f
        | \\ r
        | \\ \\
        """

    def t_string_newline(
            self,
            token):
        r' \n '
        raise ValueError(
            f'Newline within string {t}.')

    def t_string_other_characters(
            self,
            token):
        r' [^"\\]+ '

    def t_string_closing(
            self,
            token):
        r' " '
        return self._collect_string(
            token, 'STRING_LITERAL')

    def _collect_string(
            self,
            token,
            token_type):
        self.lexing_state = _INITIAL_LEXER_STATE
        # form token
        start = self._string_start
        end = token.start + len(token.value)
        value = self.input_text[start:end]
        return _lex.Token(
            symbol=token_type,
            value=value,
            start=start)

    def t_SF_WF(
            self,
            token):
        r"""
          SF_
        | WF_
        """
        return _lex.Token(
            symbol=token.value,
            value=token.value,
            start=token.start)

    def t_NAME(
            self,
            token):
        r"""
        [a-zA-Z_0-9]*
        [a-zA-Z]
        [a-zA-Z_0-9]*
        """
        if token.value in self._keywords:
            return _lex.Token(
                symbol=token.value,
                value=token.value,
                start=token.start)
        return token

    def t_BSLASH_TOKEN(
            self,
            token):
        r' \\ [a-zA-Z]+ '
        if token.value not in self._lextok:
            raise ValueError(
                f'`{token.value}` is not '
                'a known lexeme. Known lexemes '
                'of this form are in: '
                f'{self._lextok}')
        symbol = self._lextok[token.value]
        return _lex.Token(
            symbol=symbol,
            value=token.value,
            start=token.start)

    def t_FLOAT(
            self,
            token):
        r"""
        [0-9]+
        \.
        [0-9]+
        """
        return token

    def t_DECIMAL_INTEGER(
            self,
            token):
        ' [0-9]+ '
        return token

    t_omit = r' \x20* '

    def t_NEWLINE(
            self,
            token):
        r' \n+ '
        self._line_number += token.value.count('\n')

    def t_tab(
            self,
            token):
        r' \t '
        raise ValueError(
            r'TAB characters (`\t`) '
            'are unsupported.')


def _is_bslash_operator(
        lexeme):
    r"""True if `lexeme` matches `\[a-zA-Z]`."""
    return (
        lexeme.startswith('\\') and
        lexeme[1:].isalpha())


def _main():
    # read file
    args = _parse_args()
    filename = args.filename
    with open(filename, 'r') as fd:
        file_text = fd.read()
    # lex
    lexer = Lexer()
    t1 = time.perf_counter()
    tokens = lexer.parse(file_text)
    tokens = list(tokens)
    t2 = time.perf_counter()
    dt = t2 - t1
    for token in tokens:
        print(token)
    print(
        f'Success lexing file: {filename} '
        f'in {dt:2f} seconds')


def _parse_args(
        ) -> _arg.Namespace:
    """Return arguments."""
    parser = _arg.ArgumentParser()
    parser.add_argument(
        'filename',
        help='TLA+ file name to lex')
    return parser.parse_args()


if __name__ == '__main__':
    _main()



"""Installation script."""
import logging
import sys

import setuptools


PACKAGE_NAME = 'tla'
DESCRIPTION = (
    'Parser and syntax tree for TLA+, '
    'the temporal logic of actions.')
README = 'README.md'
URL = 'https://github.com/johnyf/tla'
VERSION_FILE = f'{PACKAGE_NAME}/_version.py'
VERSION = '0.0.3'
VERSION_FILE_TEXT = (
    '# This file was generated from `setup.py`\n'
    f"version = '{VERSION}'\n")
PYTHON_REQUIRES = '>=3.10'
INSTALL_REQUIRES = [
    'parstools >= 0.0.3',
    ]
TESTS_REQUIRE = [
    'pytest >= 7.4.4',
    ]
CLASSIFIERS = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Compilers',
    ]
KEYWORDS = [
    'TLA+', 'TLA', 'temporal logic of actions',
    'formal', 'specification',
    'expression', 'formula', 'module',
    'mathematics', 'theorem', 'proof',
    'parser', 'lexer', 'parsing',
    'ast', 'abstract syntax tree', 'syntax tree',
    ]


def run_setup():
    """Write version file and install package."""
    with open(README) as f:
        long_description = f.read()
    with open(VERSION_FILE, 'w') as f:
        f.write(VERSION_FILE_TEXT)
    _build_lr_parser()
    setuptools.setup(
        name=PACKAGE_NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='Ioannis Filippidis',
        author_email='jfilippidis@gmail.com',
        url=URL,
        license='BSD',
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        packages=[PACKAGE_NAME],
        package_dir={PACKAGE_NAME: PACKAGE_NAME},
        package_data={
            PACKAGE_NAME: ['tla_parser.json']},
        classifiers=CLASSIFIERS,
        keywords=KEYWORDS)


def _build_lr_parser():
    """Build the LR(1) parser state-machine."""
    if 'egg_info' in sys.argv:
        print(
            'found parameter '
            '`egg_info`, skipping '
            'building the parser')
        return
    if not _parstools_is_installed():
        return
    import tla._lre as _lr
    parser = _lr.ExprParser()
    parser.parse('1')


def _parstools_is_installed(
        ) -> bool:
    """Return `True` if `parstools` is installed."""
    try:
        import parstools
    except ImportError:
        print(
            'WARNING: package `tla` could not '
            'cache the LR(1) parser tables, '
            'during installation.')
        return False
    return True


if __name__ == '__main__':
    run_setup()



"""How to traverse syntax trees."""
import functools as _ft

import tla


_nodes = tla.make_nodes()
    # namespace of tree classes (namedtuples)


def traverse_tla_tree():
    """Traverse syntax tree to rename operators."""
    expr = r'x = 1 /\ y = 2'
    tree = tla.parse_expr(expr)
    renaming = dict(x='p', y='q')
    renamed_tree = traverse(tree, renaming)
    text = tla.pformat_ast(renamed_tree)
    print(text)


def traverse(
        tree,
        renaming:
            dict[str, str]):
    """Recursively traverse syntax tree."""
    rec = _ft.partial(
        traverse,
        renaming=renaming)
    match tree:
        case str() | bool() | None:
            return tree
        case tuple() if not hasattr(tree, 'symbol'):
            return tree
        case list():
            return list(map(rec, tree))
    match tree.symbol:
        case 'operator_application':
            op = rec(tree.operator)
            args = rec(tree.arguments)
            # operator name ?
            if tree.arguments is None:
                name = tree.operator
                op = renaming.get(name, name)
            return _nodes.OperatorApplication(
                operator=op,
                arguments=args)
        case _:
            symbol, *attrs = tree
            attrs = map(rec, attrs)
            cls = type(tree)
            return cls(symbol, *attrs)


if __name__ == '__main__':
    traverse_tla_tree()
